; ____________________________________________________________________
;|                                                                    |
;| This file is part of the project:                                  |
;|                                                                    |
;| ..::FreshLib::.. - portable, assembler library and GUI toolkit.    |
;|____________________________________________________________________|
;|                                                                    |
;|                          This file                                 |
;|                          _________                                 |
;|   Author:                                                          |
;|                                                                    |
;|   Title: StrLib.asm - OS independent string manipulation library.  |
;|                                                                    |
;|   OS:                                                              |
;|                                                                    |
;|   Notes and changes:                                               |
;|     Uses memory.asm library for memory functions.                  |
;|                                                                    |
;|                                                                    |
;|____________________________________________________________________|

; NumToStr flags
ntsSigned         = $00000000
ntsUnsigned       = $00010000
ntsZeroTerminated = $00020000
ntsFixedWidth     = $00040000

ntsBin            = $00000200
ntsQuad           = $00000400
ntsOct            = $00000800
ntsDec            = $00000a00
ntsHex            = $00001000

cStrShift     = 6
cStrChunkSize = 1 shl cStrShift

cStrChunkCount = 1024

; TChunkIndex is a dword with following format:
;  $00000000 - empty chunk
;  $pxffffff - last chunk in the string. Only $ffffff have special meaning. $fffffe for example is normal chunk number.
;              .next is unsigned 24 bit number.
;  $nxyyyyyy - first chunk in the string.       nx = not (px) - it is not NEG!
struct TChunkIndex
  .next rb 3
  .len  db ?
ends


struct TStrTableNew
  .count      dd 0      ; Count of the chunks in the string pool
  .ptrPool    dd 0      ; One big area of memory to be allocated for strings.
  .ptrIndex   dd 0      ; array of dword for every chunk in the pool: if >0 it is index of the next chunk in the string. If <0 it is occuped space in the last chunk.
  .Hint       dd ?      ; index of the fast found free chunk. From here you must start to search for the next allocation.
ends


uglobal
  _StrList    TStrTableNew
endg


initialize StrLibInit
begin
        stdcall GetMem, cStrChunkCount * cStrChunkSize
        jc      .error_init
        mov     [_StrList.ptrPool], eax

        stdcall GetMem, cStrChunkCount * 4
        jc      .error_init
        mov     [_StrList.ptrIndex], eax

        mov     [_StrList.count], cStrChunkCount
        mov     [_StrList.Hint], -1
        return

.error_init:
        int3
        return

endp




finalize StrLibDestroy
begin
        stdcall FreeMem, [_StrList.ptrPool]
        stdcall FreeMem, [_StrList.ptrIndex]
        return
endp


; ____________________________________________________________________
;| proc StrNew                                                        |
;|   Creates new dynamic string and returns the handle.               |
;| Arguments:                                                         |
;|   None                                                             |
;| Returns:                                                           |
;|   CF = 0; no error: eax contains handle of the string.             |
;|   CF = 1; error. eax = 0                                           |
;|____________________________________________________________________|
proc StrNew
begin
        push    ebx
        mov     ebx, [_StrList.ptrIndex]
        call    _search_empty_chunk
        jc      .error_new
        mov     dword [ebx+4*eax], $ffffffff    ; first chunk; last chunk; zero occupied (empty string).
        clc
        pop     ebx
        return

.error_new:
        stc
        pop     ebx
        return
endp


;__________________________________________
;
; Arguments:
;   ebx - pointer to the index.

; Returns:
;   eax - index of new empty  chunk.
;
; Returns index of a empty chunk.
; If needed, resizes memory arrays for
; chunks and for index.
; Updates _StrList as well.
;__________________________________________

proc _search_empty_chunk
begin
        push    ecx  edx

        mov     edx, [_StrList.Hint]
        mov     ecx, [_StrList.count]

.searchloop:
        inc     edx
        cmp     edx, [_StrList.count]
        jb      @f
        xor     edx, edx
@@:
        cmp     dword [ebx+4*edx], 0
        je      .free_found

        dec     ecx
        jnz     .searchloop

; the empty chunk not found. so, resize the arrays
        mov     edx, [_StrList.count]   ; that is the new string handle.

        lea     eax, [2*edx]            ; new size of the arrays. very aggressive.
        mov     [_StrList.count], eax

        shl     eax, cStrShift
        stdcall ResizeMem, [_StrList.ptrPool], eax
        jc      .finish

        mov     [_StrList.ptrPool], eax

        mov     eax, [_StrList.count]
        lea     eax, [4*eax]
        stdcall ResizeMem, [_StrList.ptrIndex], eax
        mov     [_StrList.ptrIndex], eax
        mov     ebx, eax

.free_found:
        mov     [_StrList.Hint], edx
        mov     eax, edx
.finish:
        pop     edx ecx
        return
endp



; ____________________________________________________________________
;| proc StrDel, .hString                                              |
;|   Destroys dynamic string                                          |
;| Arguments:                                                         |
;|   .hString - only handle to a string.                              |
;| Returns:                                                           |
;|   CF = 0; no error:                                                |
;|   CF = 1; error                                                    |
;|____________________________________________________________________|
proc StrDel, .hString
begin
        push    ebx edx
        mov     ebx, [_StrList.ptrIndex]
        mov     edx, [.hString]
        cmp     byte [ebx+4*edx+TChunkIndex.len], 0     ; first .len should be negative!
        jge     .invalid_handle

        mov     [_StrList.Hint], edx
        dec     [_StrList.Hint]

.del_loop:
        push    dword [ebx+4*edx]
        mov     dword [ebx+4*edx], 0    ; free chunk
        pop     edx
        and     edx, $ffffff
        cmp     edx, $ffffff
        je      .end_string

        cmp     byte [ebx+4*edx+TChunkIndex.len], 0
        jge     .del_loop                               ; every next .len should be positive.

.invalid_handle:
        stc
        pop     edx ebx
        return

.end_string:
        clc
        pop     edx ebx
        return

endp



;____________________________________________________________________
;
; proc StrCatAsciiZ, .hString, .ptrAsciiZ
;
;   Concatenates asciiz string from the memory to the given dynamic
; string.
;
; Arguments:
;   .hString - handle of a dynamic string.
;   .ptrAsciiZ - pointer to asciiz string in memory.
;
; Returns:
;   CF=1 if error.
;____________________________________________________________________

proc StrCatAsciiZ, .hString, .ptrAsciiZ
begin
        push    eax ebx ecx edx esi edi

        mov     ebx, [_StrList.ptrIndex]
        mov     eax, [.hString]

; first search to the end chunk of the string.
.endloop:
        mov     edx, eax
        mov     eax, [ebx+4*edx]
        and     eax, $ffffff
        cmp     eax, $ffffff
        jne     .endloop

        mov     eax, edx
        mov     edi, [_StrList.ptrPool]
        shl     eax, cStrShift
        add     edi, eax                 ; edi - pointer to the begin of the last chunk in string.

        movsx   ecx, byte [ebx+4*edx+TChunkIndex.len]    ; offset in the chunk.
        mov     esi, [.ptrAsciiZ]
        test    ecx, ecx
        jns     .copyloop

        not     ecx

; so, copy to the end of the chunk
.copyloop:
        cmp     ecx, cStrChunkSize
        jb      .sizeok

        call    _search_empty_chunk
        jc      .finish

; set previous chunk index.
        mov     [ebx+4*edx], eax
        mov     cl, cStrChunkSize
        cmp     edx, [.hString]
        jne     @f
        not     cl
@@:
        mov     byte [ebx+4*edx+TChunkIndex.len], cl

        mov     edi, eax
        xor     ecx, ecx  ; zero offset.
        shl     edi, cStrShift
        mov     edx, eax                ; new chunk index.
        add     edi, [_StrList.ptrPool] ; new chunk pointer

.sizeok:
        mov     al, [esi]
        inc     esi
        cmp     al, 0
        je      .endofstring
        mov     [edi+ecx], al
        inc     ecx
        jmp     .copyloop

.endofstring:
        mov     dword [ebx+4*edx+TChunkIndex.next], $00ffffff

        cmp     edx, [.hString]
        jne     @f
        not     cl
@@:
        mov     byte  [ebx+4*edx+TChunkIndex.len],  cl

.finish:
        pop     edi esi edx ecx ebx eax
        return
endp



;____________________________________________________________________
;
; proc StrCatChar, .hString, .char
;
;   Concatenates 1..4 characters to the given dynamic string.
;   Actually uses StrCatAsciiZ to do the work.
;
; Arguments:
;   .hString - handle of a dynamic string.
;   .char - dword with 1 to 4 characters.
;
; Returns:
;   CF=1 if error.
;____________________________________________________________________

proc StrCatChar, .hString, .char
.buff rd 2
begin
        push    eax
        mov     eax, [.char]
        mov     [.buff], eax
        mov     [.buff+4], 0
        lea     eax, [.buff]
        stdcall StrCatAsciiZ, [.hString], eax
        pop     eax
        return
endp




;____________________________________________________________________
;
; proc StrSplit, .hString, .pos
;
; Splits [.hString] at position [.pos].
; The left portion of the string remains in [.hString]
; The right portion forms new string that is returned by function.
; This procedure increases fragmentation of the string pool.
; In most cases it allocates one more chunk to keep the begin of
; the right part of the string.
;
; Arguments:
;   .hString - handle of string to be splitted.
;   .pos     - position for split.
; Returns:
;   CF = 1 - error.
;   CF = 0 - no error. eax contains handle of string with right part of
;            the splitted string.
;____________________________________________________________________

proc StrSplit, .hString, .pos
begin
        push    ebx ecx edx esi edi

        mov     ebx, [_StrList.ptrIndex]
        mov     edx, [.hString]
        cmp     edx, [_StrList.count]
        cmc
        jc      .finish

        movsx   eax, [ebx+4*edx+TChunkIndex.len]
        not     eax

        bt      eax, 31
        jc      .finish

        sub     [.pos], eax
        jle     .found

.lenloop:
        mov     edx, [ebx+4*edx]
        and     edx, $00ffffff
        cmp     edx, $00ffffff
        je      .exit_error

        movsx   eax, [ebx+4*edx+TChunkIndex.len]
        bt      eax, 31
        jc      .finish

        sub     [.pos], eax
        jg      .lenloop

.found:
        jl      .insertchunk

        mov     eax, [ebx+4*edx]
        and     eax, $00ffffff
        cmp     eax, $00ffffff
        je      .exit_error

        or      dword [ebx+4*edx], $00ffffff
        xor     dword [ebx+4*eax], $ff000000
        clc
        jmp     .finish

.insertchunk:
        sar     [ebx+4*edx+TChunkIndex.len], 7
        add     [.pos], eax
        mov     ecx, [.pos]   ; remaining in the left string.
        xor     [ebx+4*edx+TChunkIndex.len], cl    ; proper sign length.

        mov     esi, edx
        shl     esi, cStrShift
        add     esi, ecx        ; the offset in the pool.

        push    dword [ebx+4*edx]
        or      dword [ebx+4*edx], $00ffffff
        pop     edx
        and     edx, $00ffffff

        sub     eax, [.pos]     ; length of the remaining in the right.
        mov     ecx, eax

        call    _search_empty_chunk

        mov     [ebx+4*eax+TChunkIndex.len], cl
        xor     dword [ebx+4*eax], $ff000000    ; it is first chunk in the new string.
        or      dword [ebx+4*eax], edx          ; next chunk.

        mov     edi, eax
        shl     edi, cStrShift
        add     edi, [_StrList.ptrPool]
        add     esi, [_StrList.ptrPool]
        rep movsb
        clc
.finish:
        pop     edi esi edx ecx ebx
        return

.exit_error:
        stc
        jmp     .finish
endp



;____________________________________________________________________
;
; proc StrInsert, .hDest, .hSrc, .position
;
; Insrts [.hSrs] at [.position] inside [.hDest].
; Uses StrSplit and StrCat to do the work.
; .hStr is destroyed.
;
; Arguments:
;   .hDest - handle of destination string.
;   .hStr  - string to be inserted.
;   .position - position of insertion.
;
; Returns:
;   CF = 1 - error.
;   CF = 0 - no error.
;____________________________________________________________________

proc StrInsert, .hDest, .hSrc, .position
begin
        push    eax
        stdcall StrSplit, [.hDest], [.position]
        jc      .finish
        stdcall StrCat, [.hDest], [.hSrc]
        jc      .freeeax
        stdcall StrCat, [.hDest], eax
        jc      .strange

.finish:
        pop     eax
        return

.freeeax:
        stdcall StrCat, [.hDest], eax
        jc      .strange
        stc
        jmp     .finish

.strange:
        stdcall StrDel, eax
        stc
        jmp     .finish

endp



;____________________________________________________________________
;
; proc StrCat, .hDest, .hSrc
;
;   Concatenates two dynamic strings. The result string is [.hDest]
;   After concatenation, [.hSrc] is destroyed. If you need to keep
;   .hSrc, make copy of it with StrDup and then concatenate the
;   copy with the destination string.
;
; Arguments:
;   .hString - handle of a dynamic string.
;   .ptrAsciiZ - pointer to asciiz string in memory.
;
; Returns:
;   CF=1 if error.
;____________________________________________________________________
proc StrCat, .hDest, .hSrc
begin
        push    eax ebx ecx edx

        mov     ebx, [_StrList.ptrIndex]
        mov     eax, [.hDest]
        cmp     eax, [_StrList.count]
        cmc
        jc      .finish

        movsx   ecx, [ebx+4*eax+TChunkIndex.len]
        not     ecx
        bt      ecx, 31
        jc      .finish

.chunkloop:
        mov     edx, eax
        mov     eax, [ebx+4*edx]
        and     eax, $00ffffff
        cmp     eax, $00ffffff
        jne     .chunkloop

        mov     ecx, [.hSrc]
        cmp     ecx, [_StrList.count]
        cmc
        jc      .finish

        xor     dword [ebx+4*ecx], $ff000000
        bt      dword [ebx+4*ecx], 31
        jc      .finish

        and     dword [ebx+4*edx], $ff000000
        or      dword [ebx+4*edx], ecx

.finish:
        pop     edx ecx ebx eax
        return
endp


;____________________________________________________________________
; proc StrCompact, .hStr
;
; Compacts the string. This means, that if some chunks of the string
; are not fully used (as a result of string manipulations -
; inserts, concatenations, deletes, etc.) this function removes
; the empty spaces from the chunks and frees resulted empty chunks
; from the end of the string.
; This operation is slow if the string is fragmented and should be
; avoided without reason. Even when needed, you should call this
; procedure as later as possible, after all operations that may
; fragment the string.
;
;____________________________________________________________________
proc StrCompact, .hStr
begin


        return
endp




;____________________________________________________________________
;
; proc StrDup, .hString
;
;   Creates duplicate of the given string.
;
; Arguments:
;   .hString - handle of a dynamic string.
;
; Returns:
;   CF=1 if error.
;   CF=0 - eax = handle of the new created string.
;____________________________________________________________________
proc StrDup, .hString
.new dd ?
begin
        push    ebx ecx edx esi edi

        mov     edx, [.hString]
        mov     ebx, [_StrList.ptrIndex]
        cmp     edx, [_StrList.count]
        cmc
        jc      .finish

        movsx   ecx, [ebx+4*edx+TChunkIndex.len]
        not     ecx
        bt      ecx, 31
        jc      .finish

        call    _search_empty_chunk
        jc      .finish

        mov     [.new], eax

.chunkloop:
        mov     edi, eax
        mov     esi, edx
        shl     edi, cStrShift
        shl     esi, cStrShift
        add     edi, [_StrList.ptrPool]
        add     esi, [_StrList.ptrPool]

        mov     ecx, cStrChunkSize / 4
        rep movsd

        mov     ecx, [ebx+4*edx]
        and     ecx, $ff000000
        push    ecx

        mov     edx, [ebx+4*edx]
        and     edx, $00ffffff
        cmp     edx, $00ffffff
        je      .endofstring

        mov     ecx, eax
        call    _search_empty_chunk
        jc      .finish

        pop     dword [ebx+4*ecx]
        or      [ebx+4*ecx], eax

        jmp     .chunkloop

.endofstring:
        pop     dword [ebx+4*eax]
        or      [ebx+4*eax], edx
        mov     eax, [.new]
        clc
.finish:
        pop     edi esi edx ecx ebx
        return
endp




proc SubStr, .hString, .index, .length
begin




endp




; ____________________________________________________________________
;|                                                                    |
;| proc StrLen, .hString                                              |
;|   Returns the length of the string in bytes.                       |
;| Arguments:                                                         |
;|   .hString: pointer or handle to a string.                         |
;| Returns:                                                           |
;|    ecx - length of the strin in bytes.                             |
;|____________________________________________________________________|

proc StrLen, .hString
begin
        push    ebx edx eax

        mov     ebx, [_StrList.ptrIndex]
        mov     edx, [.hString]
        cmp     edx, [_StrList.count]
        cmc
        jc      .finish

        movsx   ecx, [ebx+4*edx+TChunkIndex.len]
        not     ecx

        bt      ecx, 31
        jc      .finish

.lenloop:
        mov     edx, [ebx+4*edx]
        and     edx, $00ffffff
        cmp     edx, $00ffffff
        je      .finish

        movsx   eax, [ebx+4*edx+TChunkIndex.len]
        bt      eax, 31
        jc      .finish

        add     ecx, eax
        jmp     .lenloop

.finish:
        pop     eax edx ebx
        return
endp


;_____________________________________________________________________
;
; proc StrFreeBuffer, .ptrBuffer
;
; Frees thestring buffer, allocated by StrPointer.
; Arguments:
;   .ptrBuffer - pointer to the buffer.
; Returns:
;   CF=1 if error. In this case, the memory is not released.
;_____________________________________________________________________

proc StrFreeBuffer, .ptrBuffer
begin
        push    eax
        mov     eax, [.ptrBuffer]
        sub     eax, 4
        stdcall FreeMem, eax
        pop     eax
        return
endp


;____________________________________________________________________________________
; proc StrPointer, .hString
;
; Extracts the string [.hString] to the allocated buffer and returns pointer
; to the allocated memory.
; Allocated memory must be freed with StrFreeBuffer.
; The buffer have dword size. The string is Zero terminated and there is guarantied
; that at least one full dword at the end is zero.
; At the [pointer - 4] is placed the length of the string in bytes.
;
; Arguments:
;   .hString - handle to a string.
; Returns:
;   esi - pointer to allocated buffer in memory.
;   CF = 1 in case of error - invalid handle of string or the memory can not be
;          allocated.
;____________________________________________________________________________________
proc StrPointer, .hString
.ptr dd ?
.len dd ?
begin
        push    eax ebx ecx edx edi

        stdcall StrLen, [.hString]
        jc      .finish

        mov     [.len], ecx
        add     ecx, 11              ; we need 8bytes more memory for length and termination
        and     cl, $fc

        stdcall GetMem, ecx
        jc      .finish

        mov     [.ptr], eax
        lea     edi, [eax+4]

        mov     ebx, [_StrList.ptrIndex]
        mov     edx, [.hString]

        cmp     dword [ebx+4*edx], 0
        je      .invalid_handle

.outerextract:
        mov     esi, edx
        shl     esi, cStrShift
        add     esi, [_StrList.ptrPool]
        movsx   ecx, [ebx+4*edx+TChunkIndex.len]
        test    ecx, ecx
        jns     .innerextract
        not     ecx

.innerextract:
        cmp     ecx, 0
        jle     .endofchunk

        movsd
        sub     ecx, 4
        jmp     .innerextract

.endofchunk:
        add     edi, ecx        ; ecx <= 0

        mov     edx, [ebx+4*edx]
        and     edx, $00ffffff
        cmp     edx, $00ffffff
        jne     .outerextract

        xor     eax, eax
        stosd                   ; terminator
        mov     esi, [.ptr]
        pushd   [.len]
        popd    [esi]
        add     esi, 4
        clc
.finish:
        pop     edi edx ecx ebx eax
        return

.invalid_handle:
        stc
        pop     edi edx ecx ebx eax
        return
endp


;_______________________________________________________________________
;
; function StrHash
;   Computes 32 bit hash value from the string.
;
; Arguments:
;   .hString - handle or pointer to the string.
;
; Return:
;   eax - 32bit hash value; FNV-1a algorithm.
;
; Changes:
;   eax
;_______________________________________________________________________
proc StrHash, .hString, .OrMask
begin
        push    edx ecx esi

        stdcall StrPtr, [.hString]
        mov     esi, eax
        mov     eax, $811C9DC5                  ; 2166136261              ; FNV offset basis
        mov     ecx, [esi-4]                    ; length of the string
        jecxz   .exit

.hashloop:
        movzx   edx, byte [esi]
        inc     esi
        or      edx, [.OrMask]
        xor     eax, edx
        imul    eax, $01000193                  ;   16777619              ; FNV prime
        dec     ecx
        jnz     .hashloop

.exit:
        pop     esi ecx edx
        return
endp
